import SwiftUI
import MapKit

struct Service: Identifiable {
    let id: Int
    let startDate: Date
    let endDate: Date
    let monday: Bool
    let tuesday: Bool
    let wednesday: Bool
    let thursday: Bool
    let friday: Bool
    let saturday: Bool
    let sunday: Bool
}

enum ServiceException {
    case active
    case inactive
}

enum RouteType: Int {
    case tram = 0
    case bus = 3
}

struct Route: Identifiable, Hashable {
    let id: Int
    let shortName: String
    let color: Color
    let textColor: Color
    let routeType: RouteType
    
    func hash(into hasher: inout Hasher) {
        hasher.combine(id)
        hasher.combine(shortName)
        hasher.combine(color)
        hasher.combine(textColor)
        hasher.combine(routeType)
    }
}

struct Stop: Identifiable, Equatable {
    let id: Int
    let name: String
    let position: CLLocationCoordinate2D
    
    static func == (lhs: Stop, rhs: Stop) -> Bool {
        return lhs.id == rhs.id
    }
}

struct StopTime: Identifiable, Equatable {
    let stopId: Int
    let secondsElapsedSinceMidnight: Int
    
    var id: String {
        "\(secondsElapsedSinceMidnight) \(stopId)"
    }
    
    var time: Date {
        let hours = secondsElapsedSinceMidnight / 3600
        let minutes = (secondsElapsedSinceMidnight % 3600) / 60
        let seconds = secondsElapsedSinceMidnight % 60
        
        var components = DateComponents()
        components.hour = hours
        components.minute = minutes
        components.second = seconds
        
        return Calendar.current.date(from: components) ?? Date()
    }
}

struct Trip: Identifiable, Equatable {
    let id: Int
    let routeId: Int
    let shapeId: Int
    let stops: [StopTime]
}

enum CustomMapElementType: Int {
    case water = 2
    case nature = 5
    case relief = 3
    case ground = 1
    case residential = 4
}

struct CustomMapPolygon: Identifiable {
    let polygon: MKPolygon
    let type: CustomMapElementType
    
    static let WATER_COLOR: Color = .blue.mix(with: Color(.systemBackground), by: 0.4)
    
    var color: Color {
        switch self.type {
        case .water:
            CustomMapPolygon.WATER_COLOR
        case .nature:
            .green.mix(with: Color(.systemBackground), by: 0.5)
        case .relief:
            .brown.mix(with: Color(.systemBackground), by: 0.7)
        case .residential:
            Color(.secondarySystemBackground)
        case .ground:
            .yellow.mix(with: Color(.secondarySystemBackground), by: 0.8)
        }
    }
    
    var id: String { "\(self.polygon.boundingMapRect)" }
}

struct Shape: Identifiable {
    let shape: MKMultiPolyline
    let color: Color
    let isTram: Bool
    
    var id: String {
        "\(color.hashValue):\(shape.boundingMapRect)"
    }
}

struct Landmark: Identifiable {
    enum Kind {
        case nature
        case museum
        case monument
    }
    
    let name: String
    let position: CLLocationCoordinate2D
    let kind: Kind?
    
    var id: String { name }
    
    init(_ name: String, _ coordinates: String, kind: Kind? = nil) {
        self.name = name
        self.kind = kind
        
        let parts = coordinates.split(separator: ",").map { Double($0.trimmingCharacters(in: .whitespaces)) }
        guard parts.count == 2, 
                let latitude = parts.first, 
                let longitude = parts.last else {
            fatalError("Invalid coordinates format. Expected format: \"latitude, longitude\"")
        }
        
        self.position = CLLocationCoordinate2D(latitude: latitude!, longitude: longitude!)
    }
    
    var color: Color {
        switch kind {
        case .nature: .green
        case .museum: .purple
        case .monument: .brown
        default: .gray
        }
    }
    
    var icon: String {
        switch kind {
        case .nature: "leaf.fill"
        case .museum: "theatermask.and.paintbrush.fill"
        case .monument: "building.columns.fill"
        default: ""
        }
    }
}

struct MapVehicle: Identifiable, Equatable {
    enum Kind {
        case tram1
        case tramOther
        case bus
    }
    
    let id: Int
    let kind: Kind
    let trip: Trip?
    let position: CLLocationCoordinate2D
    let heading: Double
    let trailerHeading: Double?
    
    var imageName: String {
        switch kind {
        case .tram1: "T1"
        case .tramOther: "T2"
        case .bus: "Bus"
        }
    }
    
    var hasTrailer: Bool { kind != .bus }
    
    static func==(lhs: MapVehicle, rhs: MapVehicle) -> Bool {
        return lhs.id == rhs.id
    }
}

struct UserSelection {
    let route: Route
    let stopTimes: [StopTime]
}

struct MapShape: Identifiable, Hashable {
    let id: Int
    let shape: MKPolyline
    let route: Route
    
    func hash(into hasher: inout Hasher) {
        hasher.combine(id)
    }
}

struct Departure: Identifiable, Equatable {
    let stop: Stop
    let route: Route
    let at: Date
    let direction: Stop
    let vehicle: MapVehicle
    
    var groupKey: String { "\(direction.id) \(route.id)" }
    
    var id: String { "\(stop.id) \(route.id) \(at.secondsElapsedSinceMidnight) \(direction.id)" }
    
    static func ==(lhs: Departure, rhs: Departure) -> Bool {
        return lhs.id == rhs.id
    }
}
