import SwiftUI
import MapKit

extension CLLocationCoordinate2D {
    func distance(to coordinate: CLLocationCoordinate2D) -> CLLocationDistance {
        let loc1 = CLLocation(latitude: self.latitude, longitude: self.longitude)
        let loc2 = CLLocation(latitude: coordinate.latitude, longitude: coordinate.longitude)
        return loc1.distance(from: loc2)
    }
    
    func isBetween(start: CLLocationCoordinate2D, end: CLLocationCoordinate2D) -> Bool {
        let crossProduct = (self.latitude - start.latitude) * (end.longitude - start.longitude) - 
        (self.longitude - start.longitude) * (end.latitude - start.latitude)
        
        if abs(crossProduct) > 0.0000001 { return false }
        
        let dotProduct = (self.longitude - start.longitude) * (end.longitude - start.longitude) + 
        (self.latitude - start.latitude) * (end.latitude - start.latitude)
        
        if dotProduct < 0 { return false }
        
        let squaredLengthBA = pow(end.longitude - start.longitude, 2) + pow(end.latitude - start.latitude, 2)
        if dotProduct > squaredLengthBA { return false }
        
        return true
    }
}

extension MKPolyline {
    func point(at progress: Double) -> (coordinate: CLLocationCoordinate2D, heading: Double)? {
        guard progress >= 0, progress <= 1, pointCount > 1 else { return nil }
        
        let totalDistance = self.totalDistance()
        let targetDistance = totalDistance * progress
        var accumulatedDistance: CLLocationDistance = 0
        
        for i in 0..<pointCount - 1 {
            let coord1 = coordinates[i]
            let coord2 = coordinates[i + 1]
            let segmentDistance = coord1.distance(to: coord2)
            
            if accumulatedDistance + segmentDistance >= targetDistance {
                let segmentProgress = (targetDistance - accumulatedDistance) / segmentDistance
                let interpolatedPoint = interpolateBetween(coord1, coord2, ratio: segmentProgress)
                
                let heading = calculateHeading(from: coord1, to: coord2)
                return (interpolatedPoint, heading)
            }
            
            accumulatedDistance += segmentDistance
        }
        
        return coordinates.last.map { ($0, 0) }
    }
    
    func intersectsCircle(center: CLLocationCoordinate2D, radius: CLLocationDistance) -> Bool {
        guard pointCount > 1 else { return false }
        
        for i in 0..<coordinates.count - 1 {
            let start = coordinates[i]
            let end = coordinates[i + 1]
            
            if start.distance(to: center) <= radius || end.distance(to: center) <= radius {
                return true
            }
            
            if segmentIntersectsCircle(from: start, to: end, center: center, radius: radius) {
                return true
            }
        }
        
        return false
    }
    
    private func segmentIntersectsCircle(from start: CLLocationCoordinate2D, to end: CLLocationCoordinate2D, center: CLLocationCoordinate2D, radius: CLLocationDistance) -> Bool {
        let projected = projectPoint(center, ontoSegmentFrom: start, to: end)
        
        let isOnSegment = projected.isBetween(start: start, end: end)
        
        let distanceToCircle = projected.distance(to: center)
        return isOnSegment && distanceToCircle <= radius
    }
    
    func subPolyline(from start: CLLocationCoordinate2D, to end: CLLocationCoordinate2D) -> MKPolyline? {
        guard let (startProjected, startIndex) = closestPointOnPolyline(to: start),
              let (endProjected, endIndex) = closestPointOnPolyline(to: end) else {
            return nil
        }
        
        let (startIdx, endIdx) = startIndex <= endIndex ? (startIndex, endIndex) : (endIndex, startIndex)
        
        var subPolylineCoordinates = [CLLocationCoordinate2D]()
        
        subPolylineCoordinates.append(startProjected)
        
        if startIdx < endIdx {
            for i in (startIdx + 1)...endIdx {
                if i < coordinates.count {
                    subPolylineCoordinates.append(coordinates[i])
                }
            }
        }
        
        subPolylineCoordinates.append(endProjected)
        
        return MKPolyline(coordinates: subPolylineCoordinates, count: subPolylineCoordinates.count)
    }
    
    private func closestPointOnPolyline(to point: CLLocationCoordinate2D) -> (CLLocationCoordinate2D, Int)? {
        guard coordinates.count > 1 else { return nil }
        
        var closestPoint: CLLocationCoordinate2D?
        var closestSegmentIndex: Int?
        var minDistance = CLLocationDistance.greatestFiniteMagnitude
        
        for i in 0..<coordinates.count - 1 {
            let segmentStart = coordinates[i]
            let segmentEnd = coordinates[i + 1]
            
            let projected = projectPoint(point, ontoSegmentFrom: segmentStart, to: segmentEnd)
            let distance = projected.distance(to: point)
            
            if distance < minDistance {
                minDistance = distance
                closestPoint = projected
                closestSegmentIndex = i
            }
        }
        
        guard let finalPoint = closestPoint, let finalIndex = closestSegmentIndex else {
            return nil
        }
        
        return (finalPoint, finalIndex)
    }
    
    private func projectPoint(_ point: CLLocationCoordinate2D, ontoSegmentFrom start: CLLocationCoordinate2D, to end: CLLocationCoordinate2D) -> CLLocationCoordinate2D {
        let startPoint = CGPoint(x: start.longitude, y: start.latitude)
        let endPoint = CGPoint(x: end.longitude, y: end.latitude)
        let targetPoint = CGPoint(x: point.longitude, y: point.latitude)
        
        let dx = endPoint.x - startPoint.x
        let dy = endPoint.y - startPoint.y
        
        let lengthSquared = dx * dx + dy * dy
        if lengthSquared == 0 { return start }
        
        let t = max(0, min(1, ((targetPoint.x - startPoint.x) * dx + (targetPoint.y - startPoint.y) * dy) / lengthSquared))
        
        let projectedX = startPoint.x + t * dx
        let projectedY = startPoint.y + t * dy
        
        return CLLocationCoordinate2D(latitude: projectedY, longitude: projectedX)
    }
    
    private func totalDistance() -> CLLocationDistance {
        var total: CLLocationDistance = 0
        for i in 0..<pointCount - 1 {
            total += coordinates[i].distance(to: coordinates[i + 1])
        }
        return total
    }
    
    private func calculateHeading(from start: CLLocationCoordinate2D, to end: CLLocationCoordinate2D) -> Double {
        let deltaLon = end.longitude - start.longitude
        let y = sin(deltaLon) * cos(end.latitude)
        let x = cos(start.latitude) * sin(end.latitude) - sin(start.latitude) * cos(end.latitude) * cos(deltaLon)
        let heading = atan2(y, x) * (180 / .pi)
        return (heading >= 0) ? heading : (heading + 360)
    }
    
    private func findClosestIndex(to target: CLLocationCoordinate2D) -> Int? {
        return coordinates.enumerated().min(by: { $0.element.distance(to: target) < $1.element.distance(to: target) })?.offset
    }
    
    private var coordinates: [CLLocationCoordinate2D] {
        var coords = [CLLocationCoordinate2D](repeating: kCLLocationCoordinate2DInvalid, count: pointCount)
        getCoordinates(&coords, range: NSRange(location: 0, length: pointCount))
        return coords
    }
    
    private func interpolateBetween(_ point1: CLLocationCoordinate2D, _ point2: CLLocationCoordinate2D, ratio: Double) -> CLLocationCoordinate2D {
        let lat = point1.latitude + (point2.latitude - point1.latitude) * ratio
        let lon = point1.longitude + (point2.longitude - point1.longitude) * ratio
        return CLLocationCoordinate2D(latitude: lat, longitude: lon)
    }
    
    private func midPoint(_ coord1: CLLocationCoordinate2D, _ coord2: CLLocationCoordinate2D) -> CLLocationCoordinate2D {
        let lat = (coord1.latitude + coord2.latitude) / 2
        let lon = (coord1.longitude + coord2.longitude) / 2
        return CLLocationCoordinate2D(latitude: lat, longitude: lon)
    }
}
