import SwiftUI
import MapKit

var routes: [Int: Route] = [:]
var trips: [Int: Trip] = [:]
var stops: [Int: Stop] = [:]
var services: [Service] = []
var serviceExceptions: [String: [Int: ServiceException]] = [:]
var shapes: [Int: MKPolyline] = [:]
var tramPolyline = Set<Int>()

struct RouteJSON: Codable {
    let short_name: String
    let route_color: String
    let route_text_color: String
    let route_type: String
    
    func toRoute(id: Int) -> Route {
        return Route(
            id: id,
            shortName: short_name,
            color: Color(hex: route_color),
            textColor: Color(hex: route_text_color.trimmingCharacters(in: .whitespacesAndNewlines)),
            routeType: RouteType(rawValue: Int(route_type) ?? 3) ?? .bus
        )
    }
}

struct TripJSON: Codable {
    let sh: Int
    let r: Int
    let s: Int
    let st: [[Int]]
    
    func toTrip(id: Int) -> Trip {
        return Trip(
            id: id,
            routeId: r,
            shapeId: sh,
            stops: st.map { StopTime(stopId: $0[0], secondsElapsedSinceMidnight: $0[1]) }
        )
    }
}

struct StopJSON: Codable {
    let coordinates: [Double]
    let name: String
    
    func toStop(id: Int) -> Stop {
        return Stop(
            id: id,
            name: name,
            position: CLLocationCoordinate2D(latitude: coordinates[0], longitude: coordinates[1])
        )
    }
}

struct ServiceJSON: Codable {
    let start_date: String
    let end_date: String
    let monday, tuesday, wednesday, thursday, friday, saturday, sunday: Bool
    
    func toService(id: Int) -> Service {
        let formatter = DateFormatter()
        formatter.dateFormat = "yyyyMMdd"
        return Service(
            id: id,
            startDate: formatter.date(from: start_date) ?? Date(),
            endDate: formatter.date(from: end_date.trimmingCharacters(in: .whitespacesAndNewlines)) ?? Date(),
            monday: monday, tuesday: tuesday, wednesday: wednesday, thursday: thursday,
            friday: friday, saturday: saturday, sunday: sunday
        )
    }
}

func loadJSON<T: Decodable, K: Hashable>(filename: String, keyType: K.Type, valueType: T.Type) async -> [K: T] {
    guard let url = Bundle.main.url(forResource: filename, withExtension: "json") else {
        print("\(filename).json not found")
        return [:]
    }
    
    do {
        let data = try Data(contentsOf: url)
        
        return try await Task.detached(priority: .background) { () -> [K: T] in
            let jsonObject = try JSONSerialization.jsonObject(with: data, options: []) as? [String: Any]
            var result: [K: T] = [:]
            
            jsonObject?.forEach { key, value in
                let castedKey: K? = {
                    if K.self == Int.self, let intKey = Int(key) {
                        return intKey as? K
                    } else if K.self == String.self {
                        return key as? K
                    }
                    return nil
                }()
                
                if let castedKey = castedKey,
                   let jsonData = try? JSONSerialization.data(withJSONObject: value),
                   let decodedValue = try? JSONDecoder().decode(T.self, from: jsonData) {
                    result[castedKey] = decodedValue
                }
            }
            
            return result
        }.value
    } catch {
        print("Error loading \(filename).json : \(error)")
        return [:]
    }
}

func mapServiceException(_ serviceDict: [String: Int]) -> [Int: ServiceException] {
    var result: [Int: ServiceException] = [:]
    
    for (key, value) in serviceDict {
        if let intKey = Int(key) {
            if value == 1 {
                result[intKey] = .active
            } else if value == 2 {
                result[intKey] = .inactive
            }
        }
    }
    
    return result
}

func loadAllData() async {
    routes = await loadJSON(filename: "routes", keyType: Int.self, valueType: RouteJSON.self)
        .mapValuesWithKey { $1.toRoute(id: $0) }
    stops = await loadJSON(filename: "stops", keyType: Int.self, valueType: StopJSON.self)
        .mapValuesWithKey { $1.toStop(id: $0) }
    services = Array(await loadJSON(filename: "calendar", keyType: Int.self, valueType: ServiceJSON.self)
        .map { $1.toService(id: $0) })
    shapes = await loadJSON(filename: "shapes", keyType: Int.self, valueType: [[Double]].self)
        .compactMapValues { coordinatesArray in
            let coordinates = coordinatesArray.map { CLLocationCoordinate2D(latitude: $0[1], longitude: $0[0]) }
            return MKPolyline(coordinates: coordinates, count: coordinates.count)
        }
    serviceExceptions = await loadJSON(filename: "calendar_dates", keyType: String.self, valueType: [String: Int].self)
        .mapValues { mapServiceException($0) }
}

func activeServiceIds() -> Set<Int> {
    let today = Date()
    let formatter = DateFormatter()
    formatter.dateFormat = "yyyyMMdd"
    let todayString = formatter.string(from: today)
    
    let calendar = Calendar.current
    let weekday = calendar.component(.weekday, from: today)
    
    let todayKey: KeyPath<Service, Bool> = {
        switch weekday {
        case 1: return \.sunday
        case 2: return \.monday
        case 3: return \.tuesday
        case 4: return \.wednesday
        case 5: return \.thursday
        case 6: return \.friday
        case 7: return \.saturday
        default: return \.monday
        }
    }()
    
    print(todayKey)
    
//    print(services)
    
    let todayServices = services
        .filter { service in
            let isWithinDateRange = (formatter.string(from: service.startDate) <= todayString) &&
            (todayString <= formatter.string(from: service.endDate))
            return isWithinDateRange && service[keyPath: todayKey]
        }
        .map { $0.id }
    
    let todayExceptions = serviceExceptions[todayString] ?? [:]
    
    let todayExceptionsActive = todayExceptions.filter { $0.value == .active }.map { $0.key }
    let todayExceptionsInactive = todayExceptions.filter { $0.value == .inactive }.map { $0.key }
    
//    print(todayServices.sorted(by: <))
//    print(todayExceptionsActive.sorted(by: <))
//    print(todayExceptionsInactive.sorted(by: <))
    return Set(todayServices + todayExceptionsActive).subtracting(todayExceptionsInactive)
}

func loadActiveTrips() async {
    let activeServices = activeServiceIds()
    trips = await loadJSON(filename: "trips", keyType: Int.self, valueType: TripJSON.self)
        .filter { _, value in
            return activeServices.contains(value.s)
        }
        .mapValuesWithKey { $1.toTrip(id: $0) }
}
