import Foundation
import MapKit
import SwiftUI

class ViewModel: ObservableObject {
    @Published var vehicles: [MapVehicle] = []
    @Published var selectedVehicle: MapVehicle? = nil
    @Published var selectedArea: CLLocationCoordinate2D? = nil
    @Published var highlightedShapes: Set<MapShape> = Set()
    @Published var nextDepartures: [Departure] = []
    
    private var lastHeadings: [Int: Double] = [:]
    
    init() {
        Task {
            await loadAllData()
            await loadActiveTrips()
        }
    }
    
    private func vehicleKind(_ route: Route) -> MapVehicle.Kind {
        if route.routeType == .bus {
            return .bus
        }
        
        return route.shortName == "L1" ? .tram1 : .tramOther
    }
    
    func highlightShapesInArea(radius: Double) {
        if let selectedTripId = selectedVehicle?.id, let trip = trips[selectedTripId], let shape = shapes[trip.shapeId], let route = routes[trip.routeId] {
            self.highlightedShapes = Set([MapShape(
                id: trip.shapeId,
                shape: shape,
                route: route
            )])
            return
        }
        
        guard let selectedArea else {
            self.highlightedShapes = Set()
            return
        }
        
        DispatchQueue.global(qos: .background).async {
            var shapesInArea = Set<MapShape>()
            var departures: [Departure] = []
            
            for vehicle in self.vehicles {
                guard let trip = trips[vehicle.id], let shape = shapes[trip.shapeId], let route = routes[trip.routeId] else { continue }
                
                for stopTime in trip.stops {
                    guard let stop = stops[stopTime.stopId] else { continue }
                    
                    if stop.position.distance(to: selectedArea) <= radius, stopTime.time.secondsElapsedSinceMidnight >= Date.now.secondsElapsedSinceMidnight {
                        guard let terminus = stops[trip.stops.last?.stopId ?? 0] else {
                            continue
                        }
                        
                        shapesInArea.insert(MapShape(
                            id: trip.shapeId,
                            shape: shape,
                            route: route
                        ))
                        
                        if stop.id != terminus.id {
                            departures.append(Departure(
                                stop: stop,
                                route: route,
                                at: stopTime.time,
                                direction: terminus,
                                vehicle: vehicle
                            ))
                        }
                        
                        break
                    }
                }
            }
            
            DispatchQueue.main.async {
                withAnimation {
                    self.highlightedShapes = shapesInArea
                    self.nextDepartures = departures.sorted { $0.at < $1.at }
                }
            }
        }
    }
    
    @MainActor
    func dismissAll() {
        selectedVehicle = nil
        selectedArea = nil
        nextDepartures = []
        highlightedShapes = Set()
    }
    
    @MainActor
    func focusVehicle(_ vehicle: MapVehicle, radius: Double) {
        selectedVehicle = vehicle
        selectedArea = nil
        nextDepartures = []
        highlightShapesInArea(radius: radius)
    }
    
    @MainActor
    func handleMapGesture(tappedLocation: CLLocationCoordinate2D, areaRadius: Double) {
        let sortedVehicles = vehicles
            .sorted { vehicle1, vehicle2 in
                let distance1 = tappedLocation.distance(to: vehicle1.position)
                let distance2 = tappedLocation.distance(to:  vehicle2.position)
                return distance1 < distance2
            }
            .filter { $0.position.distance(to: tappedLocation) < 100 }
        
        print(sortedVehicles.map { $0.id })
        
        if let closestVehicle = sortedVehicles.first {
            focusVehicle(closestVehicle, radius: areaRadius)
        } else {
            selectedVehicle = nil
            selectedArea = tappedLocation
            highlightShapesInArea(radius: areaRadius)
        }
    }
    
    func updateVehiclesPosition() {
        var updatedVehicles: [MapVehicle] = []
            
        let now = Date()
        let calendar = Calendar.current
        let secondsSinceMidnight = calendar.component(.hour, from: now) * 3600 +
        calendar.component(.minute, from: now) * 60 +
        calendar.component(.second, from: now)
        
        DispatchQueue.global(qos: .background).async {
            for vehicle in self.vehicles {
                if vehicle.hasTrailer {
                    self.lastHeadings[vehicle.id] = vehicle.heading
                }
            }
            
            for trip in trips.values {
                guard let firstStop = trip.stops.first,
                      let lastStop = trip.stops.last,
                      secondsSinceMidnight >= firstStop.secondsElapsedSinceMidnight,
                      secondsSinceMidnight <= lastStop.secondsElapsedSinceMidnight else {
                    continue
                }
                
                var previousStop: StopTime? = nil
                var nextStop: StopTime? = nil
                
                let reducedStopTimes = trip.stops.unique { $0.secondsElapsedSinceMidnight }
                
                for stopTime in reducedStopTimes {
                    if stopTime.secondsElapsedSinceMidnight > secondsSinceMidnight {
                        nextStop = stopTime
                        break
                    }
                    previousStop = stopTime
                }
                
                guard let prevStop = previousStop,
                      let nextStop = nextStop,
                      let stop1 = stops[prevStop.stopId],
                      let stop2 = stops[nextStop.stopId] else {
                    continue
                }
                
                let totalTime = Double(nextStop.secondsElapsedSinceMidnight - prevStop.secondsElapsedSinceMidnight)
                let elapsedTime = Double(secondsSinceMidnight - prevStop.secondsElapsedSinceMidnight)
                let progressRatio = elapsedTime / totalTime
                
                guard let polyline = shapes[trip.shapeId] else { continue }
                
                let subPolyline = polyline.subPolyline(from: stop1.position, to: stop2.position)
                
                guard let (position, heading) = subPolyline?.point(at: progressRatio) else { continue }
                
                guard let route = routes[trip.routeId] else { continue }
                
                if route.routeType == .tram {
                    tramPolyline.insert(trip.shapeId)
                }
                
                let vehicle = MapVehicle(
                    id: trip.id,
                    kind: self.vehicleKind(route),
                    trip: trip,
                    position: position,
                    heading: heading,
                    trailerHeading: self.lastHeadings[trip.id]
                )
                
                updatedVehicles.append(vehicle)
            }
            
            DispatchQueue.main.async {
                withAnimation {
                    self.vehicles = updatedVehicles
                }
            }
        }
    }
}
