import SwiftUI
import MapKit

struct ContentView: View {
    enum ZoomLevel {
        case minimum
        case normal
        case maximum
    }
    
    enum ContentType {
        case onboarding
        case nextDepartures
        case followVehicle
        case noDepartures
        case loading
    }
    
    @State private var polygons: [CustomMapPolygon] = []
    @State private var timeTag = Date().timeIntervalSince1970
    @State private var zoomLevel: ZoomLevel = .minimum
    @StateObject private var viewModel = ViewModel()
    
    @Environment(\.horizontalSizeClass) private var horizontalSizeClass
    
    private let TIMER_INTERVAL = 5.0
    private let AREA_RADIUS = 350.0
    private let SHEET_HEIGHT = 360.0
    
    var sortedPolygons: [CustomMapPolygon] {
        polygons.sorted { $0.type.rawValue < $1.type.rawValue }
    }
    
    var trip: Trip? {
        trips[viewModel.selectedVehicle?.id ?? 0]
    }
    
    var route: Route? {
        routes[trip?.routeId ?? 0] 
    }
    
    var tripStops: [Stop] {
        guard let trip = trip else { return [] }
        
        return trip.stops.compactMap { stops[$0.stopId] }
    }
    
    var vehicleSize: Double {
        switch zoomLevel {
        case .minimum, .normal: 10.0
        case .maximum: 20.0
        }
    }
    
    var landmarkSize: Double {
        switch zoomLevel {
        case .minimum: 24.0
        case .normal: 36.0
        case .maximum: 48.0
        }
    }
    
    var contentType: ContentType {
        if !viewModel.nextDepartures.isEmpty {
            return .nextDepartures
        } else if trip != nil, route != nil {
            return .followVehicle
        } else if viewModel.selectedArea != nil {
            return .noDepartures
        } else if viewModel.vehicles.isEmpty {
            return .loading
        } else {
            return .onboarding
        }
    }
    
    @MapContentBuilder
    func mapPolyline(_ polyline: MKPolyline, isTram: Bool) -> some MapContent {
        MapPolyline(polyline)
            .mapOverlayLevel(level: .aboveLabels)
            .stroke(
                isTram ? Color(.systemGray3) : Color(.systemGray4),
                style: StrokeStyle(
                    lineWidth: isTram ? 4 : 2,
                    lineCap: .round,
                    lineJoin: .round
                )
            )
    }
    
    var landmarksView: some MapContent {
        ForEach(LANDMARKS) { landmark in
            Annotation(landmark.name, coordinate: landmark.position) {
                LandmarkView(landmark: landmark, size: landmarkSize)
            }
            .mapOverlayLevel(level: .aboveLabels)
        }
    }
    
    var areaView: some MapContent {
        Group {
            if let center = viewModel.selectedArea {
                MapCircle(center: center, radius: AREA_RADIUS * 1.2)
                    .foregroundStyle(.green.opacity(0.1))
                    .mapOverlayLevel(level: .aboveLabels)
                MapCircle(center: center, radius: AREA_RADIUS)
                    .foregroundStyle(.green.opacity(0.3))
                    .mapOverlayLevel(level: .aboveLabels)
            }
        }
    }
    
    var activeShapes: [MapShape] {
        Array(viewModel.highlightedShapes)
            .filter { $0.id == viewModel.selectedVehicle?.trip?.shapeId }
    }
    
    var activeShapesView: some MapContent {
        Group {
            ForEach(activeShapes) { shape in
                MapPolyline(shape.shape)
                    .mapOverlayLevel(level: .aboveLabels)
                    .stroke(
                        shape.route.color,
                        style: StrokeStyle(
                            lineWidth: 4,
                            lineCap: .round,
                            lineJoin: .round
                        )
                    )
            }
        }
    }
    
    var mapBackground: some MapContent {
        Group {
            MapPolygon(MASK)
                .mapOverlayLevel(level: .aboveRoads)
                .foregroundStyle(CustomMapPolygon.WATER_COLOR)
            ForEach(sortedPolygons) { polygon in
                MapPolygon(polygon.polygon)
                    .mapOverlayLevel(level: .aboveRoads)
                    .foregroundStyle(polygon.color)
            }
            if viewModel.highlightedShapes.isEmpty {
                ForEach(shapes.sorted { _, _ in true }, id: \.key) { id, shape in
                    mapPolyline(shape, isTram: tramPolyline.contains(id))
                }
            }
        }
    }
    
    func currentVehicleSize(id: Int) -> CGFloat {
        let isSelected = viewModel.selectedVehicle?.id == id
        if zoomLevel != .minimum {
            return isSelected ? vehicleSize * 2.5 : vehicleSize
        }
        return isSelected ? vehicleSize : vehicleSize / 2
    }
    
    @ViewBuilder
    func vehicleView(_ vehicle: MapVehicle) -> some View {
        ZStack {
            Image(vehicle.imageName)
                .resizable()
                .scaledToFit()
                .frame(height: currentVehicleSize(id: vehicle.id))
                .scaleEffect(x: 1, y: vehicle.heading < 180 ? 1 : -1)
            if vehicle.hasTrailer {
                Image(vehicle.imageName)
                    .resizable()
                    .scaledToFit()
                    .frame(height: currentVehicleSize(id: vehicle.id))
                    .scaleEffect(x: -1, y: vehicle.heading < 180 ? 1 : -1)
                    .offset(x: -currentVehicleSize(id: vehicle.id) * 1.9, y: 0)
                    .rotationEffect(.degrees((vehicle.heading - (vehicle.trailerHeading ?? vehicle.heading)) * -1), anchor: .leading)
            }
        }
        .rotationEffect(.degrees(vehicle.heading - 90))
    }
    
    var departuresVehiclesId: Set<Int> {
        Set(viewModel.nextDepartures.map { $0.vehicle.id })
    }
    
    func displayBigVehicle(_ vehicle: MapVehicle) -> Bool {
        if let trip, let vehicleTrip = trips[vehicle.id] {
            return trip.routeId == vehicleTrip.routeId
        }
        
        if !viewModel.nextDepartures.isEmpty {
            return departuresVehiclesId.contains(vehicle.id)
        }
        
        return zoomLevel != .minimum
    }
    
    var vehicles: some MapContent {
        ForEach(viewModel.vehicles) { vehicle in
            Annotation("", coordinate: vehicle.position) {
                if displayBigVehicle(vehicle) {
                    vehicleView(vehicle)
                } else {
                    Circle()
                        .foregroundStyle(.gray)
                        .frame(width: 4, height: 4)
                }
            }
        }
    }
    
    var stopsView: some MapContent {
        ForEach(tripStops) { stop in
            StopDot(
                stop: stop,
                size: zoomLevel == .maximum ? 16 : 10,
                compact: zoomLevel == .minimum)
        }
    }
    
    var mapView: some View {
        MapReader { proxy in
            Map(
                initialPosition: .region(.init(center: NICE, span: SPAN))
            ) {
                mapBackground
                activeShapesView
                stopsView
                areaView
                vehicles
                landmarksView
            }
            .safeAreaInset(edge: .bottom) {
                Rectangle()
                    .opacity(0)
                    .frame(height: horizontalSizeClass == .compact ? SHEET_HEIGHT : 0)
            }
            .overlay(alignment: .bottomLeading) {
                Text("Map data from OpenStreetMap")
                    .font(.footnote)
                    .padding(.bottom, (horizontalSizeClass == .compact ? SHEET_HEIGHT : 0) + 38)
                    .padding(.horizontal, 8)
            }
            .onTapGesture { position in
                guard let tappedLocation = proxy.convert(position, from: .local) else { return }
                viewModel.handleMapGesture(tappedLocation: tappedLocation, areaRadius: AREA_RADIUS)
            }
            .mapStyle(.imagery)
            .onAppear {
                polygons = loadMapView()
                startTimer()
            }
            .onMapCameraChange(frequency: .onEnd) { cameraContext in
                let km = cameraContext.region.span.latitudeDelta * 111
                if km < 1.4 {
                    zoomLevel = .maximum
                } else if km < 4 {
                    zoomLevel = .normal
                } else {
                    zoomLevel = .minimum
                }
            }
        }
    }
    
    var detailsView: some View {
        Group {
            if contentType == .followVehicle {
                if let trip, let route {
                    StopsView(stops: stops, route: route, stopTimes: trip.stops) {
                        viewModel.dismissAll()
                    }
                    .id(viewModel.selectedVehicle?.id ?? 0)
                }
            } else if contentType == .nextDepartures {
                NextDepartures(
                    departures: viewModel.nextDepartures,
                    selectedVehicle: $viewModel.selectedVehicle
                ) {
                    viewModel.dismissAll()
                } onClick: { vehicle in
                    viewModel.focusVehicle(vehicle, radius: AREA_RADIUS)
                }
            } else if contentType == .noDepartures {
                ContentUnavailableView("No departure in this area for the moment.", systemImage: "binoculars.fill")
            } else if contentType == .loading {
                ProgressView()
            } else {
                Onboarding()
            }
        }
    }
    
    var body: some View {
        mapView
            .inspector(isPresented: .constant(true)) {
                detailsView
                    .inspectorColumnWidth(390)
                    .interactiveDismissDisabled()
                    .presentationBackgroundInteraction(.enabled)
                    .presentationDetents([.height(SHEET_HEIGHT), .large])
            }
    }
    
    private func startTimer() {
        Timer.scheduledTimer(withTimeInterval: TIMER_INTERVAL, repeats: true) { _ in
            viewModel.updateVehiclesPosition()
            timeTag = Date().timeIntervalSince1970
        }
    }
}
