import SwiftUI

struct NextDepartures: View {
    let departures: [Departure]
    @Binding var selectedVehicle: MapVehicle?
    let onClose: () -> ()
    let onClick: (MapVehicle) -> ()
    
    @Namespace private var selectionAnimation
    
    @Environment(\.colorScheme) private var colorScheme
    
    var groupedDepartures: [(key: String, departures: [Departure])] {
        Dictionary(grouping: departures) { $0.groupKey }
        .map { ($0.key, $0.value) }
        .sorted {
            let a = $0.departures.first?.at
            let b = $1.departures.first?.at
            
            if let a, let b, a != b { return a < b }
            
            return $0.key < $1.key
        }
    }
    
    @ViewBuilder
    func routeIndicator(_ route: Route) -> some View {
        Text(route.shortName)
            .font(.footnote)
            .padding(.horizontal, 4)
            .padding(.vertical, 1)
            .background(route.color)
            .foregroundColor(route.textColor)
            .cornerRadius(5)
    }
    
    @ViewBuilder
    func imageView(_ vehicle: MapVehicle) -> some View {
        if vehicle.hasTrailer {
            HStack(spacing: -4) {
                Image(vehicle.imageName)
                    .resizable()
                    .scaledToFit()
                    .scaleEffect(x: -1)
                    .rotationEffect(.degrees(-18), anchor: .trailing)
                    .shadow(radius: 2, x: 3, y: 2)
                Image(vehicle.imageName)
                    .resizable()
                    .scaledToFit()
                    .shadow(radius: 2, x: 3, y: 2)
            }
            .rotationEffect(.degrees(25))
            .matchedGeometryEffect(id: vehicle.id, in: selectionAnimation)
        } else {
            Image(vehicle.imageName)
                .resizable()
                .scaledToFit()
                .rotationEffect(.degrees(25))
                .shadow(radius: 2, x: 3, y: 2)
                .matchedGeometryEffect(id: vehicle.id, in: selectionAnimation)
        }
    }
    
    @ViewBuilder
    func textView(_ firstDeparture: Departure, _ otherDepartures: [Departure]) -> some View {
        HStack {
            VStack(alignment: .leading) {
                Text(firstDeparture.direction.name)
                    .font(.title2)
                HStack {
                    routeIndicator(firstDeparture.route)
                    Text("From \(firstDeparture.stop.name)")
                }
                    .font(.footnote)
                if !otherDepartures.isEmpty {
                    Text("Next at " + otherDepartures
                        .map { $0.at.formatted(date: .omitted, time: .shortened) }
                        .joined(separator: ", "))
                    .font(.footnote)
                }
            }
            Spacer()
            Text(firstDeparture.at.formatted(date: .omitted, time: .shortened))
                .font(.title2)
        }
//        .matchedGeometryEffect(id: firstDeparture.id, in: selectionAnimation)
    }
    
    @ViewBuilder
    func departureView(_ departures: [Departure]) -> some View {
        let otherDepartures = Array(departures.dropFirst())
        
        if let firstDeparture = departures.first {
            if selectedVehicle?.id == firstDeparture.vehicle.id {
                VStack {
                    imageView(firstDeparture.vehicle)
                        .frame(height: 46)
                        .padding(.vertical, 32)
                    textView(firstDeparture, otherDepartures)
                }
                .padding()
                .cornerRadius(12)
                .background {
                    RoundedRectangle(cornerRadius: 22)
                        .foregroundStyle(.background
                            .shadow(.inner(color: firstDeparture.route.color.opacity(colorScheme == .light ? 0.2 : 0.5), radius: 14, x: -4, y: 2))
                        )
                }
                .overlay(alignment: .topTrailing) {
                    Button {
                        onClick(firstDeparture.vehicle)
                    } label: {
                        Label("Follow", systemImage: "location.fill")
                    }
                    .tint(firstDeparture.route.color)
                    .buttonStyle(.bordered)
                    .buttonBorderShape(.capsule)
                    .padding()
                }
            } else {
                HStack(spacing: 24) {
                    imageView(firstDeparture.vehicle)
                        .frame(width: 56)
                    textView(firstDeparture, otherDepartures)
                }
                .padding(.vertical, 16)
                .padding(.horizontal)
            }
        }
    }
    
    var listView: some View {
        VStack {
            ForEach(groupedDepartures, id: \.key) { group in
                departureView(group.departures)
                    .contentShape(Rectangle())
                    .onTapGesture {
                        withAnimation {
                            selectedVehicle = group.departures.first?.vehicle
                        }
                    }
            }
        }
        .padding(12)
    }
    
    var scrollView: some View {
        ScrollView {
            VStack(alignment: .leading) {
                listView
                Disclaimers(include: [.gtfs, .timezone])
                    .padding()
            }
        }
    }
    
    var body: some View {
        NavigationStack {
            ZStack {
                scrollView
            }
            .toolbar {
                ToolbarItem(placement: .topBarTrailing) {
                    CloseButton(color: .gray) {
                        onClose()
                    }
                }
            }
            .navigationTitle("Next departures")
            .navigationBarTitleDisplayMode(.inline)
        }
        .onAppear {
            selectedVehicle = groupedDepartures.first?.departures.first?.vehicle
        }
        .onChange(of: departures) {
            selectedVehicle = groupedDepartures.first?.departures.first?.vehicle
        }
    }
}

#Preview {
    NextDepartures(
        departures: [
            Departure(
                stop: Stop(id: 8, name: "Henri Sappia", position: NICE),
                route: Route(id: 1, shortName: "L1", color: .red, textColor: .white, routeType: .tram),
                at: .now.addingTimeInterval(60 * 1),
                direction: Stop(id: 2, name: "Hopital Pasteur", position: NICE),
                vehicle: MapVehicle(id: 1, kind: .tram1, trip: nil, position: NICE, heading: 0, trailerHeading: nil)),
            Departure(
                stop: Stop(id: 1, name: "La Bornala", position: NICE),
                route: Route(id: 12, shortName: "12+", color: .pink, textColor: .white, routeType: .bus),
                at: .now.addingTimeInterval(60 * 5),
                direction: Stop(id: 2, name: "Magnan", position: NICE),
                vehicle: MapVehicle(id: 2, kind: .bus, trip: nil, position: NICE, heading: 0, trailerHeading: nil)),
            Departure(
                stop: Stop(id: 1, name: "La Bornala", position: NICE),
                route: Route(id: 12, shortName: "12+", color: .pink, textColor: .white, routeType: .bus),
                at: .now.addingTimeInterval(60 * 12),
                direction: Stop(id: 2, name: "Magnan", position: NICE),
                vehicle: MapVehicle(id: 3, kind: .bus, trip: nil, position: NICE, heading: 0, trailerHeading: nil)),
            Departure(
                stop: Stop(id: 6, name: "Jean Medecin", position: NICE),
                route: Route(id: 2, shortName: "L2", color: .blue, textColor: .white, routeType: .tram),
                at: .now.addingTimeInterval(60 * 30),
                direction: Stop(id: 2, name: "Aéroport T2", position: NICE),
                vehicle: MapVehicle(id: 4, kind: .tramOther, trip: nil, position: NICE, heading: 0, trailerHeading: nil))
        ],
        selectedVehicle: .constant(nil)
    ) {
        
    } onClick: { vehicle in
        
    } 
}
