import SwiftUI

struct StopsView: View {
    let stops: [Int: Stop]
    let route: Route
    let stopTimes: [StopTime]
    let onClose: () -> ()
    
    let RECTANGLE_WIDTH = 12.0
    let RECTANGLE_PADDING = 9.5
    
    @State private var remainingStops: [StopTime] = []
    
    private func startTimer() {
        Timer.scheduledTimer(withTimeInterval: 5.0, repeats: true) { _ in
            withAnimation(.timingCurve(0.85, 0.06, 0.52, 0.9, duration: 1.5)) {
                updateRemainingStops()
            }
        }
    }
    
    func updateRemainingStops() {
        let secondsSinceMidnight = Date().secondsElapsedSinceMidnight
        remainingStops = stopTimes.filter { $0.secondsElapsedSinceMidnight > secondsSinceMidnight }
    }
    
    var direction: String? {
        guard let lastStop = stopTimes.last,
              let stop = stops[lastStop.stopId] else { return nil }
        
        return stop.name
    }
    
    var routeIndicator: some View {
        Text(route.shortName)
            .font(.headline)
            .padding(.horizontal, 6)
            .padding(.vertical, 3)
            .background(route.color)
            .foregroundColor(route.textColor)
            .cornerRadius(5)
    }
    
    var stopsList: some View {
        VStack(alignment: .leading, spacing: 0) {
            HStack(alignment: .bottom, spacing: 22) {
                Rectangle()
                    .foregroundStyle(route.color)
                    .frame(width: RECTANGLE_WIDTH, height: 58)
                    .padding(.leading, RECTANGLE_PADDING)
                Text("Next stop")
                    .font(.caption)
                    .opacity(0.6)
            }
            ForEach(remainingStops) { stopTime in
                if let stop = stops[stopTime.stopId] {
                    ZStack(alignment: .leading) {
                        HStack(spacing: 22) {
                            if stopTimes.last == stopTime {
                                Rectangle()
                                    .foregroundStyle(route.color)
                                    .frame(width: RECTANGLE_WIDTH, height: 41)
                                    .padding(.leading, RECTANGLE_PADDING)
                                    .padding(.bottom, 41)
                            } else {
                                Rectangle()
                                    .foregroundStyle(route.color)
                                    .frame(width: RECTANGLE_WIDTH, height: 82)
                                    .padding(.leading, RECTANGLE_PADDING)
                            }
                            HStack {
                                Text(stop.name)
                                    .font(.title2)
                                Text(stopTime.time.formatted(date: .omitted, time: .shortened))
                                    .font(.headline)
                                    .opacity(0.6)
                            }
                        }
                        BlinkingDotView(blink: stopTime == remainingStops.first)
                            .id("\(stopTime.id) \(remainingStops.count)")
                    }
                }
            }
        }
    }
    
    var header: some View {
        HStack {
            routeIndicator
            if let direction {
                Text(direction)
                    .font(.title)
                    .foregroundStyle(route.textColor)
            }
        }
    }
    
    var body: some View {
        ScrollView {
            ZStack(alignment: .topLeading) {
                LinearGradient(
                    gradient: Gradient(
                        colors: [
                            route.color.mix(with: Color(.clear), by: 0.1),
                            route.color.mix(with: Color(.clear), by: 0.3),
                            route.color.mix(with: Color(.clear), by: 0.8),
                            Color(.clear)
                        ]
                    ),
                    startPoint: .top,
                    endPoint: .bottom
                )
                .ignoresSafeArea()
                .frame(height: 160)
                
                VStack(alignment: .leading, spacing: 2) {
                    header
                    if remainingStops.isEmpty {
                        Text("This service has ended.")
                            .font(.headline)
                            .padding(.top)
                    } else {
                        stopsList
                        Disclaimers(include: [.gtfs, .timezone])
                            .padding(.top)
                    }
                }
                .padding()
                
                HStack {
                    Spacer()
                    CloseButton(color: route.textColor) {
                        onClose()
                    }
                }
                .padding()
            }
            .frame(maxWidth: .infinity, maxHeight: .infinity)
            .presentationDragIndicator(.visible)
        }
        .onAppear {
            startTimer()
            updateRemainingStops()
        }
    }
}

#Preview {
    Text("")
        .inspector(isPresented: .constant(true)) {
            StopsView(
                stops: [
                    1: Stop(id: 1, name: "Foo", position: NICE),
                    2: Stop(id: 2, name: "Bar", position: NICE),
                    3: Stop(id: 3, name: "Baz", position: NICE)
                ],
                route: Route(
                    id: 4,
                    shortName: "L1",
                    color: .red,
                    textColor: .white,
                    routeType: .tram
                ),
                stopTimes: [
                    StopTime(stopId: 1, secondsElapsedSinceMidnight: 3600 * 22 + 60 * 30),
                    StopTime(stopId: 2, secondsElapsedSinceMidnight: 3600 * 22 + 60 * 35),
                    StopTime(stopId: 3, secondsElapsedSinceMidnight: 3600 * 22 + 60 * 48),
                ]
            ) {
                
            }.presentationDetents([.medium])
        }
}
